import uuid
from dataclasses import dataclass, field
from typing import List, Any, Optional, Dict

import pinecone
from pinecone import QueryResponse

from autochain.models.base import BaseLanguageModel
from autochain.tools.base import Tool
from autochain.tools.internal_search.base_search_tool import BaseSearchTool


@dataclass
class PineconeDoc:
    doc: str
    vector: List[float] = None
    id: str = field(default_factory=lambda: str(uuid.uuid1()))


class PineconeSearch(Tool, BaseSearchTool):
    """
    Use Pinecone as the internal search tool
    """

    docs: List[PineconeDoc]
    index_name: str = "index"
    index: Optional[Any] = None
    dimension: int = 8
    metric: str = "euclidean"
    encoder: BaseLanguageModel = None  # such as OpenAIAdaEncoder
    id2doc: Dict[str, str] = {}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        pinecone.create_index(
            self.index_name, dimension=self.dimension, metric=self.metric
        )
        self.index = pinecone.Index(self.index_name)

        self.add_docs(self.docs)

    def _encode(self, doc: PineconeDoc) -> None:
        if not doc.vector and self.encoder:
            # TODO: encoder over batches
            doc.vector = self.encoder.encode([doc.doc]).embeddings[0]

    def _run(
        self,
        query: str,
        top_k: int = 2,
        include_values: bool = False,
        *args: Any,
        **kwargs: Any,
    ) -> str:
        def _format_output(query_response: QueryResponse) -> str:
            """Only return the document since they are likely to be passed to prompt"""
            documents = query_response.get("matches", [])
            if len(documents) == 0:
                return ""

            return "\n".join(
                [
                    f"Doc {i}: {self.id2doc[doc['id']]}"
                    for i, doc in enumerate(documents)
                ]
            )

        encoding = self.encoder.encode([query]).embeddings[0]

        response: QueryResponse = self.index.query(
            vector=encoding, top_k=top_k, include_values=include_values
        )
        return _format_output(response)

    def add_docs(self, docs: List[PineconeDoc], **kwargs):
        if not len(docs):
            return

        for doc in docs:
            self._encode(doc)
            self.id2doc[doc.id] = doc.doc

        self.index.upsert([(d.id, d.vector) for d in docs])

    def clear_index(self):
        pinecone.delete_index(self.index_name)
        pinecone.create_index(
            self.index_name, dimension=self.dimension, metric=self.metric
        )
        self.index = pinecone.Index(self.index_name)



import argparse
import logging
import os
from typing import Optional, Dict, Any

from colorama import Style


def print_with_color(text: str, color: str):
    print(color + text)
    print(Style.RESET_ALL)


def get_from_dict_or_env(
    data: Dict[str, Any], key: str, env_key: str, default: Optional[str] = None
) -> str:
    """Get a value from a dictionary or an environment variable."""
    if key in data and data[key]:
        return data[key]
    else:
        return get_from_env(key, env_key, default=default)


def get_from_env(key: str, env_key: str, default: Optional[str] = None) -> str:
    """Get a value from a dictionary or an environment variable."""
    if env_key in os.environ and os.environ[env_key]:
        return os.environ[env_key]
    elif default is not None:
        return default
    else:
        raise ValueError(
            f"Did not find {key}, please add an environment variable"
            f" `{env_key}` which contains it, or pass"
            f"  `{key}` as a named parameter."
        )


def get_args():
    """Adding arguments for running test interactively or setting verbosity"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--interact",
        "-i",
        action="store_true",
        help="if run interactively",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="if show detailed contents, such as intermediate results and prompts",
    )
    args = parser.parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.INFO)

    return args



from typing import List, Optional

from langchain.agents import AgentType, initialize_agent
from langchain.base_language import BaseLanguageModel as LCBaseLanguageModel
from langchain.chat_models import ChatOpenAI as LCChatOpenAIModel
from langchain.memory import ConversationBufferMemory as LCConversationBufferMemory
from langchain.schema import BaseMemory as LCBaseMemory
from langchain.tools import Tool as LCTool
from autochain.chain.langchain_wrapper_chain import LangChainWrapperChain
from autochain.workflows_evaluation.langchain_eval.custom_langchain_output_parser import (
    CustomConvoOutputParser,
)


def create_langchain_from_test(
    tools: List[LCTool],
    agent_type: AgentType,
    memory: Optional[LCBaseMemory] = None,
    llm: Optional[LCBaseLanguageModel] = None,
    **kwargs,
):
    """
    Create LangChainWrapperChain by instantiating LangChain agent
    Args:
        tools: list of langchain tool
        agent_type: LangChain AgentType
        memory: LangChain memory
        llm: LangChain language model

    Returns:
        LangChainWrapperChain
    """
    llm = llm or LCChatOpenAIModel(temperature=0)
    memory = memory or LCConversationBufferMemory(memory_key="chat_history")

    # Created a more lenient output parser to walk around fragility of LangChain Agent
    kwargs["output_parser"] = CustomConvoOutputParser()

    langchain = initialize_agent(
        tools, llm, agent=agent_type, verbose=True, memory=memory, agent_kwargs=kwargs
    )

    return LangChainWrapperChain(langchain=langchain)



import enum
from abc import abstractmethod
from typing import List, Any, Dict

from pydantic import BaseModel, Field


class MessageType(enum.Enum):
    UserMessage = enum.auto()
    AIMessage = enum.auto()
    SystemMessage = enum.auto()
    FunctionMessage = enum.auto()


class BaseMessage(BaseModel):
    """Message object."""

    content: str
    additional_kwargs: dict = Field(default_factory=dict)

    @property
    @abstractmethod
    def type(self) -> str:
        """Type of the message, used for serialization."""


class UserMessage(BaseMessage):
    """Type of message that is spoken by the human."""

    example: bool = False

    @property
    def type(self) -> str:
        """Type of the message, used for serialization."""
        return "user"


class AIMessage(BaseMessage):
    """Type of message that is spoken by the AI."""

    example: bool = False
    function_call: Dict[str, Any] = {}

    @property
    def type(self) -> str:
        """Type of the message, used for serialization."""
        return "ai"


class SystemMessage(BaseMessage):
    """Type of message that is a system message."""

    @property
    def type(self) -> str:
        """Type of the message, used for serialization."""
        return "system"


class FunctionMessage(BaseMessage):
    """Type of message that is a function message."""

    name: str

    @property
    def type(self) -> str:
        """Type of the message, used for serialization."""
        return "function"


class ChatMessageHistory(BaseModel):
    messages: List[BaseMessage] = []

    def save_message(self, message: str, message_type: MessageType, **kwargs):
        if message_type == MessageType.AIMessage:
            self.messages.append(AIMessage(content=message))
        elif message_type == MessageType.UserMessage:
            self.messages.append(UserMessage(content=message))
        elif message_type == MessageType.FunctionMessage:
            self.messages.append(FunctionMessage(content=message, name=kwargs["name"]))
        elif message_type == MessageType.SystemMessage:
            self.messages.append(SystemMessage(content=message))

    def format_message(self):
        string_messages = []
        if len(self.messages) > 0:
            for m in self.messages:
                if isinstance(m, UserMessage):
                    role = "User"
                elif isinstance(m, AIMessage):
                    role = "Assistant"
                elif isinstance(m, SystemMessage):
                    role = "System"
                else:
                    continue
                string_messages.append(f"{role}: {m.content}")
            return "\n".join(string_messages) + "\n"
        return ""

    def get_latest_user_message(self) -> UserMessage:
        for message in reversed(self.messages):
            if isinstance(message, UserMessage):
                return message
        return UserMessage(content="n/a")

    def clear(self) -> None:
        self.messages = []



import os.path
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Tuple, Any, Dict

import pandas as pd
from colorama import Fore

from autochain.agent.message import UserMessage
from autochain.chain import constants
from autochain.chain.base_chain import BaseChain
from autochain.models.base import Generation
from autochain.models.chat_openai import ChatOpenAI
from autochain.tools.base import Tool
from autochain.utils import print_with_color
from autochain.workflows_evaluation.test_utils import parse_evaluation_response


@dataclass
class TestCase:
    """Standardized data class for each test case for BastTest"""

    test_name: str = ""
    user_context: str = ""
    expected_outcome: str = ""


class BaseTest(ABC):
    @property
    @abstractmethod
    def chain(self) -> BaseChain:
        """Chain to test with, which support run(<user_query>) function"""

    @property
    @abstractmethod
    def tools(self) -> List[Tool]:
        """Workflow policy"""

    @property
    @abstractmethod
    def test_cases(self) -> List[TestCase]:
        """"""


class WorkflowTester:
    def __init__(self, tests: List[BaseTest], output_dir: str):
        self.chain = None
        self.tests = tests
        self.output_dir = output_dir
        self.llm = ChatOpenAI(temperature=0)

    def test_each_case(self, test_case: TestCase):
        self.chain.memory.clear()

        conversation_history = []
        user_query = ""
        conversation_end = False
        max_turn = 8
        response = {}
        while not conversation_end and len(conversation_history) < max_turn:
            if not conversation_end:
                user_query = self.get_next_user_query(
                    conversation_history, test_case.user_context
                )
                conversation_history.append(("user", user_query))
                print_with_color(f">> User: {user_query}", Fore.GREEN)

            response: Dict[str, Any] = self.chain.run(user_query)

            agent_message = response["message"]
            conversation_history.append(("assistant", agent_message))
            print_with_color(f">> Assistant: {agent_message}", Fore.GREEN)

            conversation_end = self.determine_if_conversation_ends(agent_message)

        is_agent_helpful = self.determine_if_agent_solved_problem(
            conversation_history, test_case.expected_outcome
        )
        return conversation_history, is_agent_helpful, response

    def run_test(self, test):
        test_results = []
        self.chain = test.chain
        for i, test_case in enumerate(test.test_cases):
            print(
                f"========== Start running test case: {test_case.test_name} ==========\n"
            )
            conversation_history, is_agent_helpful, last_response = self.test_each_case(
                test_case
            )
            test_results.append(
                {
                    "test_name": test_case.test_name,
                    "conversation_history": [
                        f"{user_type}: {message}"
                        for user_type, message, in conversation_history
                    ],
                    "num_turns": len(conversation_history),
                    "expected_outcome": test_case.expected_outcome,
                    "is_agent_helpful": is_agent_helpful,
                    "actions_took": [
                        {
                            "tool": action.tool,
                            "tool_input": action.tool_input,
                            "tool_output": action.tool_output,
                        }
                        for action in last_response[constants.INTERMEDIATE_STEPS]
                    ],
                }
            )

        df = pd.DataFrame(test_results)
        os.makedirs(self.output_dir, exist_ok=True)
        df.to_json(
            os.path.join(self.output_dir, f"{test.__class__.__name__}.jsonl"),
            lines=True,
            orient="records",
        )

    def run_all_tests(self):
        for test in self.tests:
            self.run_test(test)

    def run_interactive(self):
        test = self.tests[0]
        self.chain = test.chain
        self.chain.memory.clear()

        while True:
            user_query = input(">> User: ")
            response = self.chain.run(user_query)["message"]
            print_with_color(f">> Assistant: {response}", Fore.GREEN)

    def determine_if_conversation_ends(self, last_utterance: str) -> bool:
        messages = [
            UserMessage(
                content=f"""The most recent reply from assistant
assistant: "{last_utterance}"
Has assistant finish assisting the user? Answer with yes or no"""
            ),
        ]
        output: Generation = self.llm.generate(messages=messages).generations[0]

        if "yes" in output.message.content.lower():
            # finish assisting; conversation should end
            return True
        else:
            # not yet finished; conversation should continue
            return False

    def get_next_user_query(
        self, conversation_history: List[Tuple[str, str]], user_context: str
    ) -> str:
        messages = []
        conversation = ""

        for user_type, utterance in conversation_history:
            conversation += f"{user_type}: {utterance}\n"

        conversation += "user: "

        messages.append(
            UserMessage(
                content=f"""You are a user with access to the following context information about yourself. 
Based on previous conversation, write the message to assistant to help you with goal described 
in context without asking repetitive questions.
Replies 'Thank you' if the goal is achieved.
If you are not sure about how to answer, respond with "hand off to agent".
Context:
"{user_context}"

Previous conversation:
{conversation}"""
            )
        )

        output: Generation = self.llm.generate(
            messages=messages, stop=[".", "?"]
        ).generations[0]
        return output.message.content

    def determine_if_agent_solved_problem(
        self, conversation_history: List[Tuple[str, str]], expected_outcome: str
    ) -> Dict[str, str]:
        messages = []
        conversation = ""
        for user_type, utterance in conversation_history:
            conversation += f"{user_type}: {utterance}\n"

        messages.append(
            UserMessage(
                content=f"""You are an admin for assistant and check if assistant meets the expected outcome based on previous conversation.
 
Previous conversation:
{conversation}

Expected outcome is "{expected_outcome}"
Does conversation reach the expected outcome for user? answer in JSON format
{{
    "reason": "explain step by step if conversation reaches the expected outcome",
    "rating": "rating from 1 to 5; 1 for not meeting the expected outcome at all, 5 for completely meeting the expected outcome",
}}"""
            )
        )

        output: Generation = self.llm.generate(messages=messages).generations[0]
        return parse_evaluation_response(output.message)



"""
This is an example implementation of long term memory and retrieve using query
It contains three memory stores.
conversation_history stores all the messages including FunctionMessage between assistant and agent,
long_term_memory stores a collection of ChromaDoc (or would be modified use other vectory db)
kv_memory: stores anything else as kv pairs
"""
from typing import Any, Optional

from autochain.agent.message import ChatMessageHistory, MessageType
from autochain.memory.base import BaseMemory
from autochain.tools.internal_search.base_search_tool import BaseSearchTool
from autochain.tools.internal_search.chromadb_tool import ChromaDBSearch, ChromaDoc
from autochain.tools.internal_search.pinecone_tool import PineconeSearch, PineconeDoc


class LongTermMemory(BaseMemory):
    """Buffer for storing conversation memory and an in-memory kv store."""

    conversation_history = ChatMessageHistory()
    kv_memory = {}
    long_term_memory: BaseSearchTool = None

    class Config:
        keep_untouched = (
            ChromaDBSearch,
            PineconeSearch,
        )

    def load_memory(
        self,
        key: Optional[str] = None,
        default: Optional[Any] = None,
        top_k: int = 1,
        **kwargs
    ) -> Any:
        """Return history buffer by key or all memories."""
        if key in self.kv_memory:
            return self.kv_memory[key]

        # else try to retrieve from long term memory
        result = self.long_term_memory.run({"query": key, "top_k": top_k})
        return result or default

    def load_conversation(self, **kwargs) -> ChatMessageHistory:
        """Return history buffer and format it into a conversational string format."""
        return self.conversation_history

    def save_memory(self, key: str, value: Any) -> None:
        if (
            isinstance(value, list)
            and len(value) > 0
            and (isinstance(value[0], ChromaDoc) or isinstance(value[0], PineconeDoc))
        ):
            self.long_term_memory.add_docs(docs=value)
        elif key:
            self.kv_memory[key] = value

    def save_conversation(
        self, message: str, message_type: MessageType, **kwargs
    ) -> None:
        """Save context from this conversation to buffer."""
        self.conversation_history.save_message(
            message=message, message_type=message_type, **kwargs
        )

    def clear(self) -> None:
        """Clear memory contents."""
        self.conversation_history.clear()
        self.long_term_memory.clear_index()
        self.kv_memory = {}



from string import Template
from typing import Any, List

from pydantic import BaseModel, Extra

from autochain.agent.message import BaseMessage, UserMessage


class JSONPromptTemplate(BaseModel):
    """
    Format prompt with string Template and dictionary of variables
    """

    template: Template
    """The prompt template."""

    input_variables: List[str]
    """A list of the names of the variables the prompt template expects."""

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        arbitrary_types_allowed = True

    def format_prompt(self, **kwargs: Any) -> List[BaseMessage]:
        variables = {v: "" for v in self.input_variables}
        variables.update(kwargs)
        prompt = self.template.substitute(**variables)
        return [UserMessage(content=prompt)]



from langchain.agents import AgentType
from langchain.tools import Tool as LCTool

from autochain.workflows_evaluation.base_test import BaseTest, TestCase, WorkflowTester
from autochain.utils import get_args
from autochain.workflows_evaluation.langchain_eval.langchain_test_utils import (
    create_langchain_from_test,
)


def check_order_status(order_id: str, **kwargs):
    """Returns order information as a dictionary, where order_status can be "shipped" or "not_shipped" """
    if "123" in order_id:
        return {
            "status_code": 200,
            "order_id": "123",
            "order_status": "shipped",
            "tracking_url": "example.com/123",
            "shipping_address": "123 ivy street san francisco ca",
        }
    elif "456" in order_id:
        return {
            "status_code": 200,
            "order_id": "456",
            "order_status": "not_shipped",
            "tracking_url": "example.com/456",
            "shipping_address": "234 spear street san francisco ca",
        }
    else:
        return {"status_code": 400, "message": "order not found"}


def change_shipping_address(order_id: str, new_address: str = "", **kwargs):
    """Changes the shipping address for unshipped orders. Requires the order_id and the new_address inputs"""
    return {
        "status_code": 200,
        "order_id": order_id,
        "shipping_address": new_address,
    }


class TestChangeShippingAddressWithLC(BaseTest):
    prompt = """When a customer requests to change their shipping address, verify the order status in the system based on order id.
If the order has not yet shipped, update the shipping address as requested and confirm with the customer that it has been updated. 
If the order has already shipped, inform them that it is not possible to change the shipping address at this stage and provide assistance on how to proceed with exchanges, by following instructions at example.com/returns.

TOOLS:
------

Assistant has access to the following tools:
"""

    tools = [
        LCTool(
            name="check order status",
            func=check_order_status,
            description="""This function checks the order status based on order_id
Input args: order_id: non-empty str""",
        ),
        LCTool(
            name="change shipping address",
            func=change_shipping_address,
            description="""This function change the shipping address based on provided 
order_id and new_address 
Input args: order_id: non-empty str, new_address: non-empty str""",
        ),
    ]

    test_cases = [
        TestCase(
            test_name="change shipping address",
            user_context="order id is 456. the new address is 234 spear st, "
            "san francisco",
            expected_outcome="found order status and changed shipping address",
        ),
    ]

    chain = create_langchain_from_test(
        tools=tools,
        agent_type=AgentType.CONVERSATIONAL_REACT_DESCRIPTION,
        prefix=prompt,
    )


if __name__ == "__main__":
    tests = WorkflowTester(
        tests=[TestChangeShippingAddressWithLC()], output_dir="./test_results"
    )

    args = get_args()
    if args.interact:
        tests.run_interactive()
    else:
        tests.run_all_tests()



import json

from autochain.agent.openai_funtions_agent.openai_functions_agent import (
    OpenAIFunctionsAgent,
)
from autochain.chain.chain import Chain
from autochain.memory.buffer_memory import BufferMemory
from autochain.models.chat_openai import ChatOpenAI
from autochain.tools.base import Tool
from autochain.utils import get_args

# Set logging level
_ = get_args()


def get_current_weather(location: str, unit: str = "fahrenheit"):
    """Get the current weather in a given location"""
    weather_info = {
        "location": location,
        "temperature": "72",
        "unit": unit,
        "forecast": ["sunny", "windy"],
    }
    return json.dumps(weather_info)


tools = [
    Tool(
        name="get_current_weather",
        func=get_current_weather,
        description="""Get the current weather in a given location""",
    )
]

memory = BufferMemory()
llm = ChatOpenAI(temperature=0)
agent = OpenAIFunctionsAgent.from_llm_and_tools(llm=llm, tools=tools)
chain = Chain(agent=agent, memory=memory)

# example
user_query = "What's the weather today?"
print(f">> User: {user_query}")
print(f">> Assistant: {chain.run(user_query)['message']}")
next_user_query = "Boston"
print(f">> User: {next_user_query}")
print(f">> Assistant: {chain.run(next_user_query)['message']}")



import uuid
from dataclasses import dataclass, field
from typing import List, Any, Dict, Optional

import chromadb
from chromadb.api import QueryResult
from pydantic import Extra

from autochain.tools.base import Tool
from autochain.tools.internal_search.base_search_tool import BaseSearchTool


@dataclass
class ChromaDoc:
    doc: str
    metadata: Dict[str, Any]
    id: str = field(default_factory=lambda: str(uuid.uuid1()))


class ChromaDBSearch(Tool, BaseSearchTool):
    """
    Use ChromaDB as internal search tool
    """

    collection_name: str = "index"
    collection: Optional[Any] = None

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        arbitrary_types_allowed = True

    def __init__(self, docs: List[ChromaDoc], **kwargs):
        super().__init__(**kwargs)
        client = chromadb.Client()

        collection = client.create_collection(self.collection_name)
        self.collection = collection

        # Add docs to the collection. Can also update and delete. Row-based API coming soon!
        self.add_docs(docs=docs)

    def _run(
        self,
        query: str,
        top_k: int = 2,
        *args: Any,
        **kwargs: Any,
    ) -> str:
        def _format_output(query_result: QueryResult) -> str:
            """Only return the document since they are likely to be passed to prompt"""
            documents = query_result.get("documents", [])
            if len(documents) == 0:
                return ""

            docs = documents[0]
            return "\n".join([f"Doc {i}: {doc}" for i, doc in enumerate(docs)])

        result = self.collection.query(
            query_texts=[query],
            n_results=top_k,
        )
        return _format_output(result)

    def add_docs(self, docs: List[ChromaDoc], **kwargs):
        """Add a list of documents to collection"""
        if docs:
            self.collection.add(
                documents=[d.doc for d in docs],
                # we embed for you, or bring your own
                metadatas=[d.metadata for d in docs],
                # filter on arbitrary metadata!
                ids=[d.id for d in docs],  # must be unique for each doc
            )

    def clear_index(self):
        self.collection.delete()



import os
from typing import List, Optional, Any, Dict

from pydantic import root_validator

from autochain.tools.base import Tool

from autochain.agent.message import BaseMessage

from autochain.models.base import BaseLanguageModel, LLMResult, EmbeddingResult


class OpenAIAdaEncoder(BaseLanguageModel):
    """
    Text encoder using OpenAI Model
    """

    client: Any  #: :meta private:
    model_name: str = "text-embedding-ada-002"

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that api key and python package exists in environment."""
        openai_api_key = os.environ["OPENAI_API_KEY"]
        try:
            import openai

        except ImportError:
            raise ValueError(
                "Could not import openai python package. "
                "Please install it with `pip install openai`."
            )
        openai.api_key = openai_api_key
        try:
            values["client"] = openai.Embedding
        except AttributeError:
            raise ValueError(
                "`openai` has no `ChatCompletion` attribute, this is likely "
                "due to an old version of the openai package. Try upgrading it "
                "with `pip install --upgrade openai`."
            )
        return values

    def generate(
        self,
        messages: List[BaseMessage],
        functions: Optional[List[Tool]] = None,
        stop: Optional[List[str]] = None,
    ) -> LLMResult:
        pass

    def encode(self, texts: List[str]) -> EmbeddingResult:
        def _format_response(texts, resp) -> EmbeddingResult:
            embeddings = [d.get("embedding") for d in resp.get("data", [])]
            return EmbeddingResult(texts=texts, embeddings=embeddings)

        params: Dict[str, Any] = {
            "model": self.model_name,
            "input": texts,
            **self._default_params,
        }

        response = self.generate_with_retry(**params)
        return _format_response(texts=texts, resp=response)



import json
from typing import Union

from autochain.agent.message import AIMessage
from autochain.agent.structs import AgentAction, AgentFinish, AgentOutputParser


class OpenAIFunctionOutputParser(AgentOutputParser):
    def parse(self, message: AIMessage) -> Union[AgentAction, AgentFinish]:
        if message.function_call:
            action_name = message.function_call["name"]
            action_args = json.loads(message.function_call["arguments"])

            return AgentAction(
                tool=action_name,
                tool_input=action_args,
                model_response=message.content,
            )
        else:
            return AgentFinish(message=message.content, log=message.content)



from abc import abstractmethod
from typing import Any, List

from pydantic import Extra, BaseModel


class BaseSearchTool(BaseModel):
    class Config:
        """Configuration for this pydantic object."""

    extra = Extra.forbid
    arbitrary_types_allowed = True

    @abstractmethod
    def _run(
        self,
        query: str,
        top_k: int = 2,
        *args: Any,
        **kwargs: Any,
    ) -> str:
        raise NotImplementedError

    @abstractmethod
    def add_docs(self, docs: List[Any], **kwargs):
        raise NotImplementedError

    @abstractmethod
    def clear_index(self):
        raise NotImplementedError



import pickle
from typing import Any, Optional, Dict

from autochain.agent.message import (
    ChatMessageHistory,
    MessageType,
    BaseMessage,
    AIMessage,
    UserMessage,
    FunctionMessage,
    SystemMessage,
)
from autochain.memory.base import BaseMemory
from redis import Redis

from autochain.memory.constants import ONE_HOUR


class RedisMemory(BaseMemory):
    """Store conversation info in redis memory."""

    expire_time: int = ONE_HOUR
    redis_key_prefix: str
    redis_client: Redis

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def load_memory(
        self, key: Optional[str] = None, default: Optional[Any] = None, **kwargs
    ) -> Any:
        """Get the key's corresponding value from redis."""
        if not key.startswith(self.redis_key_prefix):
            key = self.redis_key_prefix + f":{key}"
        pickled = self.redis_client.get(key)
        if not pickled:
            return default
        return pickle.loads(pickled)

    def load_conversation(self, **kwargs: Dict[str, Any]) -> ChatMessageHistory:
        """Return chat message history."""
        redis_key = self.redis_key_prefix + f":{ChatMessageHistory.__name__}"
        return ChatMessageHistory(messages=self.load_memory(redis_key, []))

    def save_memory(self, key: str, value: Any) -> None:
        """Save the key value pair to redis."""
        if not key.startswith(self.redis_key_prefix):
            key = self.redis_key_prefix + f":{key}"
        pickled = pickle.dumps(value)
        self.redis_client.set(key, pickled, ex=self.expire_time)

    def save_conversation(
        self, message: str, message_type: MessageType, **kwargs
    ) -> None:
        """Save context from this conversation to redis."""
        redis_key = self.redis_key_prefix + f":{ChatMessageHistory.__name__}"
        pickled = self.redis_client.get(redis_key)
        if pickled:
            messages: list[BaseMessage] = pickle.loads(pickled)
        else:
            messages = []
        if message_type == MessageType.AIMessage:
            messages.append(AIMessage(content=message))
        elif message_type == MessageType.UserMessage:
            messages.append(UserMessage(content=message))
        elif message_type == MessageType.FunctionMessage:
            messages.append(FunctionMessage(content=message, name=kwargs["name"]))
        elif message_type == MessageType.SystemMessage:
            messages.append(SystemMessage(content=message))
        else:
            raise ValueError(f"Unsupported message type: {message_type}")
        self.save_memory(redis_key, messages)

    def clear(self) -> None:
        """Clear redis memory."""
        for key in self.redis_client.keys(f"{self.redis_key_prefix}:*"):
            self.redis_client.delete(key)



import json
from typing import Union

from colorama import Fore

from autochain.agent.message import BaseMessage
from autochain.agent.structs import AgentAction, AgentFinish, AgentOutputParser
from autochain.errors import OutputParserException
from autochain.utils import print_with_color


class ConvoJSONOutputParser(AgentOutputParser):
    def parse(self, message: BaseMessage) -> Union[AgentAction, AgentFinish]:
        response = self.load_json_output(message)

        action_name = response.get("tool", {}).get("name")
        action_args = response.get("tool", {}).get("args")

        if (
            "no" in response.get("thoughts", {}).get("need_use_tool").lower().strip()
            or not action_name
        ):
            output_message = response.get("response")
            if output_message:
                return AgentFinish(message=response.get("response"), log=output_message)
            else:
                return AgentFinish(
                    message="Sorry, i don't understand", log=output_message
                )

        return AgentAction(
            tool=action_name,
            tool_input=action_args,
            model_response=response.get("response", ""),
        )

    def parse_clarification(
        self, message: BaseMessage, agent_action: AgentAction
    ) -> Union[AgentAction, AgentFinish]:
        response = self.load_json_output(message)

        has_arg_value = response.get("has_arg_value", "")
        clarifying_question = response.get("clarifying_question", "")

        if "no" in has_arg_value.lower() and clarifying_question:
            return AgentFinish(message=clarifying_question, log=clarifying_question)
        else:
            return agent_action



"""OpenAI chat wrapper."""
from __future__ import annotations

import logging
from typing import Any, Dict, List, Optional, Tuple, Union

import torch
from pydantic import Field
from transformers import pipeline, AutoModelForCausalLM, AutoTokenizer

from autochain.agent.message import (
    BaseMessage,
    AIMessage,
)
from autochain.models.base import (
    LLMResult,
    Generation,
    BaseLanguageModel,
)
from autochain.tools.base import Tool

logger = logging.getLogger(__name__)


class HuggingFaceTextGenerationModel(BaseLanguageModel):
    """Huggingface model that supports text-generation task

    Example:
    .. code-block:: python

        from autochain.models.huggingface_text_generation_model import HuggingFaceTextGenerationModel
        llm = HuggingFaceTextGenerationModel(model_name="mosaicml/mpt-7b", model_kwargs={"trust_remote_code":True})
    """
    model_name: str = "gpt2"
    """Model name to use. GPT2 is only for demostration purpose. It does not work well for task 
    planning"""
    temperature: float = 0
    """What sampling temperature to use."""
    model_kwargs: Dict[str, Any] = Field(default_factory=dict)
    """Holds any model parameters valid for `create` call not explicitly specified."""

    tokenizer_kwargs: Dict[str, Any] = Field(default_factory=dict)
    """Holds any model parameters valid for creating tokenizer."""

    request_timeout: Optional[Union[float, Tuple[float, float]]] = None
    """Timeout for requests to OpenAI completion API. Default is 600 seconds."""
    max_retries: int = 6
    # TODO: support streaming
    # """Maximum number of retries to make when generating."""
    # streaming: bool = False
    # """Whether to stream the results or not."""
    # n: int = 1
    """Number of chat completions to generate for each prompt."""
    max_tokens: Optional[int] = 512
    """Maximum number of tokens to generate."""

    default_stop_tokens: List[str] = ["."]
    """Model will generate tokens up to the number of max token, so it would be good to have 
    default stop token"""

    model: Optional[AutoModelForCausalLM]
    tokenizer: Optional[AutoTokenizer]

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if torch.cuda.is_available():
            self.model_kwargs["device_map"] = "auto"

        self.tokenizer = AutoTokenizer.from_pretrained(
            self.model_name, **self.tokenizer_kwargs
        )

    def generate(
        self,
        messages: List[BaseMessage],
        functions: Optional[List[Tool]] = None,
        stop: Optional[List[str]] = None,
    ) -> LLMResult:
        generator = pipeline(
            task="text-generation",
            model=self.model_name,
            tokenizer=self.tokenizer,
            max_new_tokens=self.max_tokens,
            temperature=self.temperature,
            **self.model_kwargs,
        )

        prompt = self._construct_prompt_from_message(messages)
        generation = generator(prompt, do_sample=False)

        return self._create_llm_result(generation=generation, prompt=prompt, stop=stop)

    @staticmethod
    def _construct_prompt_from_message(messages: List[BaseMessage]):
        prompt = ""
        for msg in messages:
            prompt += msg.content
        return prompt

    @staticmethod
    def _enforce_stop_tokens(text: str, stop: List[str]) -> str:
        """Cut off the text as soon as any stop words occur."""
        first_index = len(text)
        for s in stop:
            if s in text:
                first_index = min(text.index(s), first_index)

        return text[:first_index].strip()

    def _create_llm_result(
        self, generation: List[Dict[str, Any]], prompt: str, stop: List[str]
    ) -> LLMResult:
        text = generation[0]["generated_text"][len(prompt):]
        if self.max_tokens:
            token_ids = self.tokenizer.encode(text)[: self.max_tokens]
            text = self.tokenizer.decode(token_ids)

        # it is better to have a default stop token so model does not always generate to max
        # sequence length
        stop = stop or self.default_stop_tokens
        text = self._enforce_stop_tokens(text=text, stop=stop)

        return LLMResult(
            generations=[Generation(message=AIMessage(content=text))],
            llm_output={
                "token_usage": len(text.split()),
                "model_name": self.model_name,
            },
        )



from __future__ import annotations

import logging
from string import Template
from typing import Any, Dict, List, Optional, Union

from colorama import Fore

from autochain.agent.base_agent import BaseAgent
from autochain.agent.conversational_agent.output_parser import ConvoJSONOutputParser
from autochain.agent.conversational_agent.prompt import (
    CLARIFYING_QUESTION_PROMPT_TEMPLATE,
    PLANNING_PROMPT_TEMPLATE,
    SHOULD_ANSWER_PROMPT_TEMPLATE,
    FIX_TOOL_INPUT_PROMPT_TEMPLATE,
)
from autochain.agent.message import BaseMessage, ChatMessageHistory, UserMessage
from autochain.agent.prompt_formatter import JSONPromptTemplate
from autochain.agent.structs import AgentAction, AgentFinish
from autochain.models.base import BaseLanguageModel, Generation
from autochain.tools.base import Tool
from autochain.utils import print_with_color

logger = logging.getLogger(__name__)


class ConversationalAgent(BaseAgent):
    """
    Conversational agent who can use tools available to make a conversation by following
    the conversational planning prompt
    """

    output_parser: ConvoJSONOutputParser = ConvoJSONOutputParser()
    llm: BaseLanguageModel = None
    prompt_template: JSONPromptTemplate = None
    allowed_tools: Dict[str, Tool] = {}
    tools: List[Tool] = []

    # Optionally you could set a prompt for this conversational agent or directly update the prompt
    prompt: str = ""

    @classmethod
    def from_llm_and_tools(
        cls,
        llm: BaseLanguageModel,
        tools: Optional[List[Tool]] = None,
        output_parser: Optional[ConvoJSONOutputParser] = None,
        prompt_template: str = PLANNING_PROMPT_TEMPLATE,
        input_variables: Optional[List[str]] = None,
        prompt: str = "",
        **kwargs: Any,
    ) -> ConversationalAgent:
        """Construct an agent from an LLM and tools."""
        tools = tools or []

        template = cls.get_prompt_template(
            template=prompt_template,
            input_variables=input_variables,
        )

        allowed_tools = {tool.name: tool for tool in tools}
        _output_parser = output_parser or ConvoJSONOutputParser()
        return cls(
            llm=llm,
            allowed_tools=allowed_tools,
            output_parser=_output_parser,
            prompt_template=template,
            tools=tools,
            prompt=prompt,
            **kwargs,
        )

    def should_answer(
        self,
        should_answer_prompt_template: str = SHOULD_ANSWER_PROMPT_TEMPLATE,
        **kwargs,
    ) -> Optional[AgentFinish]:
        """Determine if agent should continue to answer user questions based on the latest user
        query"""
        if "history" not in kwargs or not kwargs["history"]:
            return None

        history = kwargs.pop("history")
        inputs = {
            "history": history.format_message(),
            **kwargs,
        }

        def _parse_response(res: str):
            if "yes" in res.lower():
                return AgentFinish(
                    message="Thank your for contacting",
                    log="Thank your for contacting",
                )
            else:
                return None

        prompt = Template(should_answer_prompt_template).substitute(**inputs)
        response = (
            self.llm.generate([UserMessage(content=prompt)])
            .generations[0]
            .message.content
        )
        return _parse_response(response)

    @staticmethod
    def format_prompt(
        template: JSONPromptTemplate,
        intermediate_steps: List[AgentAction],
        **kwargs: Any,
    ) -> List[BaseMessage]:
        def _construct_scratchpad(
            actions: List[AgentAction],
        ) -> Union[str, List[BaseMessage]]:
            scratchpad = ""
            for action in actions:
                scratchpad += action.response
            return scratchpad

        """Create the planning inputs for the LLMChain from intermediate steps."""
        thoughts = _construct_scratchpad(intermediate_steps)
        new_inputs = {"agent_scratchpad": thoughts}
        full_inputs = {**kwargs, **new_inputs}
        prompt = template.format_prompt(**full_inputs)
        return prompt

    @staticmethod
    def get_prompt_template(
        template: str = "",
        input_variables: Optional[List[str]] = None,
    ) -> JSONPromptTemplate:
        """Create prompt in the style of the zero shot agent.

        Args:
            template: message to be injected between prefix and suffix.
            input_variables: List of input variables the final prompt will expect.

        Returns:
            A PromptTemplate with the template assembled from the pieces here.
        """
        template = Template(template)

        if input_variables is None:
            input_variables = ["input", "agent_scratchpad"]
        return JSONPromptTemplate(template=template, input_variables=input_variables)

    def plan(
        self,
        history: ChatMessageHistory,
        intermediate_steps: List[AgentAction],
        **kwargs: Any,
    ) -> Union[AgentAction, AgentFinish]:
        """
        Plan the next step. either taking an action with AgentAction or respond to user with AgentFinish
        Args:
            history: entire chat conversation between user and agent including the latest query
            intermediate_steps: List of AgentAction that has been performed with outputs
            **kwargs: key value pairs from chain, which contains query and other stored memories

        Returns:
            AgentAction or AgentFinish
        """
        print_with_color("Planning", Fore.LIGHTYELLOW_EX)
        tool_names = ", ".join([tool.name for tool in self.tools])
        tool_strings = "\n\n".join(
            [f"> {tool.name}: \n{tool.description}" for tool in self.tools]
        )
        inputs = {
            "tool_names": tool_names,
            "tools": tool_strings,
            "history": history.format_message(),
            "prompt": self.prompt,
            **kwargs,
        }
        final_prompt = self.format_prompt(
            self.prompt_template, intermediate_steps, **inputs
        )
        logger.info(f"\nPlanning Input: {final_prompt[0].content} \n")

        full_output: Generation = self.llm.generate(final_prompt).generations[0]
        agent_output: Union[AgentAction, AgentFinish] = self.output_parser.parse(
            full_output.message
        )

        print(f"Planning output: \n{repr(full_output.message.content)}", Fore.YELLOW)
        if isinstance(agent_output, AgentAction):
            print_with_color(
                f"Plan to take action '{agent_output.tool}'", Fore.LIGHTYELLOW_EX
            )

        return agent_output

    def clarify_args_for_agent_action(
        self,
        agent_action: AgentAction,
        history: ChatMessageHistory,
        intermediate_steps: List[AgentAction],
        **kwargs: Any,
    ):
        """
        Ask clarifying question if needed. When agent is about to perform an action, we could
        use this function with different prompt to ask clarifying question for input if needed.
        Sometimes the planning response would already have the clarifying question, but we found
        it is more precise if there is a different prompt just for clarifying args

        Args:
            agent_action: agent action about to take
            history: conversation history including the latest query
            intermediate_steps: list of agent action taken so far
            **kwargs:

        Returns:
            Either a clarifying question (AgentFinish) or take the planned action (AgentAction)
        """
        print_with_color("Deciding if need clarification", Fore.LIGHTYELLOW_EX)
        if not self.allowed_tools.get(agent_action.tool):
            return agent_action
        else:
            inputs = {
                "tool_name": agent_action.tool,
                "tool_desp": self.allowed_tools.get(agent_action.tool).description,
                "history": history.format_message(),
                **kwargs,
            }

            clarifying_template = self.get_prompt_template(
                template=CLARIFYING_QUESTION_PROMPT_TEMPLATE
            )

            final_prompt = self.format_prompt(
                clarifying_template, intermediate_steps, **inputs
            )
            logger.info(f"\nClarification inputs: {final_prompt[0].content}")
            full_output: Generation = self.llm.generate(final_prompt).generations[0]
            print(f"Clarification outputs: {repr(full_output.message.content)}")
            return self.output_parser.parse_clarification(
                full_output.message, agent_action=agent_action
            )

    def fix_action_input(
        self, tool: Tool, action: AgentAction, error: str
    ) -> AgentAction:
        """If the tool failed due to error, what should be the fix for inputs"""
        prompt = FIX_TOOL_INPUT_PROMPT_TEMPLATE.format(
            tool_description=tool.description, inputs=action.tool_input, error=error
        )

        logger.info(f"\nFixing tool input prompt: {prompt}")
        messages = UserMessage(content=prompt)
        output = self.llm.generate([messages]).generations[0]
        new_tool_inputs = self.output_parser.load_json_output(output.message)

        logger.info(f"\nFixed tool output: {new_tool_inputs}")
        new_action = AgentAction(tool=action.tool, tool_input=new_tool_inputs)
        return new_action



from typing import Any

from autochain.tools.base import Tool


class HandOffToAgent(Tool):
    name = "Hand off"
    description = "Hand off to a human agent"
    handoff_msg = "Let me hand you off to an agent now"

    def _run(self, *args: Any, **kwargs: Any) -> str:
        return self.handoff_msg



class OutputParserException(Exception):
    """Exception that output parsers should raise to signify a parsing error.

    This exists to differentiate parsing errors from other code or execution errors
    that also may arise inside the output parser. OutputParserExceptions will be
    available to catch and handle in ways to fix the parsing error, while other
    errors will be raised.
    """

    pass


class ToolRunningError(Exception):
    """Exception when tool fails to run"""

    def __init__(self, message):
        self.message = message



from autochain.agent.openai_funtions_agent.openai_functions_agent import (
    OpenAIFunctionsAgent,
)
from autochain.models.chat_openai import ChatOpenAI
from autochain.tools.base import Tool
from autochain.workflows_evaluation.base_test import BaseTest, TestCase, WorkflowTester
from autochain.workflows_evaluation.test_utils import (
    create_chain_from_test,
)
from autochain.utils import get_args


def check_order_status(order_id: str, **kwargs):
    """Returns order information as a dictionary, where order_status can be "shipped" or "not_shipped" """
    if order_id == "123":
        return {
            "status_code": 200,
            "order_id": "123",
            "order_status": "shipped",
            "tracking_url": "example.com/123",
            "shipping_address": "123 ivy street san francisco ca",
        }
    elif order_id == "456":
        return {
            "status_code": 200,
            "order_id": "456",
            "order_status": "not_shipped",
            "tracking_url": "example.com/456",
            "shipping_address": "234 spear street san francisco ca",
        }
    else:
        return {"status_code": 400, "message": "order not found"}


def change_shipping_address(order_id: str, new_address: str, **kwargs):
    """Changes the shipping address for unshipped orders. Requires the order_id and the new_address inputs"""
    return {
        "status_code": 200,
        "order_id": order_id,
        "shipping_address": new_address,
    }


class TestChangeShippingAddressWithFunctionCalling(BaseTest):
    prompt = """When a customer requests to change their shipping address, verify the order status in the system based on order id.
If the order has not yet shipped, update the shipping address as requested and confirm with the customer that it has been updated. 
If the order has already shipped, inform them that it is not possible to change the shipping address at this stage and provide assistance on how to proceed with exchanges, by following instructions at example.com/returns.
"""

    tools = [
        Tool(
            func=check_order_status,
            description="""This function checks the order status based on order_id
    Input args: order_id: non-empty str""",
        ),
        Tool(
            func=change_shipping_address,
            description="""This function change the shipping address based on provided 
    order_id and new_address 
    Input args: order_id: non-empty str, new_address: non-empty str""",
        ),
    ]

    test_cases = [
        TestCase(
            test_name="change shipping address",
            user_context="order id is 456. the new address is 234 spear st, "
            "san francisco",
            expected_outcome="found order status and changed shipping address",
        ),
        TestCase(
            test_name="failed changing shipping address, no order id",
            user_context="don't know about order id. the new address is 234 spear st, san francisco",
            expected_outcome="cannot find the order status, failed to change shipping "
            "address",
        ),
        TestCase(
            test_name="failed changing shipping address, shipped item",
            user_context="order id is 123. the new address is 234 spear st, "
            "san francisco",
            expected_outcome="inform user cannot change shipping address and hand off to "
            "agent",
        ),
    ]

    llm = ChatOpenAI(temperature=0)
    chain = create_chain_from_test(
        tools=tools, agent_cls=OpenAIFunctionsAgent, llm=llm, prompt=prompt
    )


if __name__ == "__main__":
    tester = WorkflowTester(
        tests=[TestChangeShippingAddressWithFunctionCalling()],
        output_dir="./test_results",
    )

    args = get_args()
    if args.interact:
        tester.run_interactive()
    else:
        tester.run_all_tests()



from langchain.agents import AgentType
from langchain.tools import Tool as LCTool

from autochain.workflows_evaluation.base_test import BaseTest, TestCase, WorkflowTester
from autochain.utils import get_args
from autochain.workflows_evaluation.langchain_eval.langchain_test_utils import (
    create_langchain_from_test,
)


def get_item_spec(item_name: str, **kwargs):
    if "toy" in item_name.lower():
        return {"name": "toy bear", "color": "red", "age_group": "1-5 years old"}
    elif "printer" in item_name.lower():
        return {
            "name": "Wireless Printer",
            "printer_type": "Printer, Scanner, Copier",
            "color_print_speed": "5.5 page per minute",
            "mono_print_speed": "7.5 page per minute",
        }
    else:
        return {}


def search_image_path_for_item(item_name: str):
    if "toy" in item_name.lower():
        return "images/toy.png"
    elif "printer" in item_name.lower():
        return "images/awesome_printer.png"
    else:
        return ""


class TestGenerateAdsWithLC(BaseTest):
    prompt = """Your goals is helping user to generate an advertisement for user requested 
product and find relevant image path for the item.
You would first clarify what product you would write advertisement for and what are the key 
points should be included in the ads.
Based on item name, you could get its specifications that can be used in advertisement.
Then, you need to search and include an image path for the item at the bottom of advertisement. 
You could find relevant images path with tool provided and search of relevant image using query.
Generate advertisement with image path.

TOOLS:
------

Assistant has access to the following tools:
"""

    tools = [
        LCTool(
            name="get item spec",
            func=get_item_spec,
            description="""This function get item spec by searching for item name
Input args: item_name: non-empty str""",
        ),
        LCTool(
            name="search image path for item",
            func=search_image_path_for_item,
            description="""This function retrieves relevant image path for a given search query
Input args: item_name: str""",
        ),
    ]

    test_cases = [
        TestCase(
            test_name="ads for toy bear",
            user_context="Write me an advertisement for toy bear; item name is 'toy bear'. it is "
            "cute and made in USA, they should be "
            "included in the ads. Ads should include image",
            expected_outcome="generate an advertisement for toy bear and mentions it is cute. "
            "Also ads should include an image path",
        ),
        TestCase(
            test_name="printer ads",
            user_context="write me an advertisement for printer; item name is 'good printer'. "
            "printer is used and in good condition. "
            "Ads should include image",
            expected_outcome="generate an advertisement for wireless printer and mentions it is "
            "wireless, can be used as scanner and is used. Also ads should "
            "include an image path",
        ),
    ]

    chain = create_langchain_from_test(
        tools=tools,
        agent_type=AgentType.CONVERSATIONAL_REACT_DESCRIPTION,
        prefix=prompt,
    )


if __name__ == "__main__":
    tests = WorkflowTester(tests=[TestGenerateAdsWithLC()], output_dir="./test_results")

    args = get_args()
    if args.interact:
        tests.run_interactive()
    else:
        tests.run_all_tests()



from __future__ import annotations

import logging
from typing import Any, Dict, List, Optional, Union

from colorama import Fore

from autochain.agent.base_agent import BaseAgent
from autochain.agent.message import ChatMessageHistory, SystemMessage
from autochain.agent.openai_funtions_agent.output_parser import (
    OpenAIFunctionOutputParser,
)
from autochain.agent.structs import AgentAction, AgentFinish
from autochain.models.base import BaseLanguageModel, Generation
from autochain.tools.base import Tool
from autochain.utils import print_with_color

logger = logging.getLogger(__name__)


class OpenAIFunctionsAgent(BaseAgent):
    """
    Agent supports function calling natively in OpenAI, which leverage function message to
    determine which tool should be used
    When tool is not selected, responds just like conversational agent
    Tool descriptions are generated from typing from the tool
    """

    llm: BaseLanguageModel = None
    allowed_tools: Dict[str, Tool] = {}
    tools: List[Tool] = []
    prompt: Optional[str] = None

    @classmethod
    def from_llm_and_tools(
        cls,
        llm: BaseLanguageModel,
        tools: Optional[List[Tool]] = None,
        output_parser: Optional[OpenAIFunctionOutputParser] = None,
        prompt: str = None,
        **kwargs: Any,
    ) -> OpenAIFunctionsAgent:
        tools = tools or []

        allowed_tools = {tool.name: tool for tool in tools}
        _output_parser = output_parser or OpenAIFunctionOutputParser()
        return cls(
            llm=llm,
            allowed_tools=allowed_tools,
            output_parser=_output_parser,
            tools=tools,
            prompt=prompt,
            **kwargs,
        )

    def plan(
        self,
        history: ChatMessageHistory,
        intermediate_steps: List[AgentAction],
        **kwargs: Any,
    ) -> Union[AgentAction, AgentFinish]:
        print_with_color("Planning", Fore.LIGHTYELLOW_EX)

        final_messages = []
        if self.prompt:
            final_messages.append(SystemMessage(content=self.prompt))
        final_messages += history.messages

        logger.info(f"\nPlanning Input: {[m.content for m in final_messages]} \n")
        full_output: Generation = self.llm.generate(
            final_messages, self.tools
        ).generations[0]

        agent_output: Union[AgentAction, AgentFinish] = self.output_parser.parse(
            full_output.message
        )
        print(
            f"Planning output: \nmessage content: {repr(full_output.message.content)}; "
            f"function_call: "
            f"{repr(full_output.message.function_call)}",
            Fore.YELLOW,
        )
        if isinstance(agent_output, AgentAction):
            print_with_color(
                f"Plan to take action '{agent_output.tool}'", Fore.LIGHTYELLOW_EX
            )

        return agent_output



"""Wrapper of LangChain to follow the same chain interface"""
from typing import Any, Dict, List, Optional

from langchain.chains.base import Chain as LangChain
from langchain.schema import BaseMemory

from autochain.agent.structs import AgentFinish, AgentAction
from autochain.chain.base_chain import BaseChain
from autochain.tools.base import Tool


class LangChainWrapperChain(BaseChain):
    """
    Wrapper chain instantiate from LangChain's Chain object to match AutoChain interface
    """

    langchain: LangChain = None
    memory: Optional[BaseMemory] = None

    def __init__(self, langchain: LangChain, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.langchain = langchain
        self.memory = self.langchain.memory

    def run(
        self,
        user_query: str,
        **kwargs,
    ) -> Dict[str, Any]:
        response_msg: str = self.langchain.run(user_query)
        agent_finish = AgentFinish(message=response_msg, log="")
        return agent_finish.format_output()

    def take_next_step(
        self,
        name_to_tool_map: Dict[str, Tool],
        inputs: Dict[str, str],
    ) -> (AgentFinish, AgentAction):
        pass



"""
Base interface that all chains should implement
"""
import logging
import time
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Any, Dict, List, Optional

from autochain.agent.base_agent import BaseAgent
from autochain.agent.message import ChatMessageHistory, MessageType
from autochain.agent.structs import AgentAction, AgentFinish
from autochain.chain import constants
from autochain.memory.base import BaseMemory
from autochain.tools.base import Tool
from pydantic import BaseModel

logger = logging.getLogger(__name__)


class BaseChain(BaseModel, ABC):
    """
    Base interface that all chains should implement.
    Chain will standardize inputs and outputs, the main entry pointy is the run function.
    """

    agent: Optional[BaseAgent] = None
    memory: Optional[BaseMemory] = None
    last_query: str = ""
    max_iterations: Optional[int] = 15
    max_execution_time: Optional[float] = None

    def prep_inputs(self, user_query: str) -> Dict[str, str]:
        """Load conversation history from memory and prep inputs."""
        inputs = {
            constants.CONVERSATION_HISTORY: ChatMessageHistory(),
            constants.INTERMEDIATE_STEPS: [],
        }
        if self.memory is not None:
            intermediate_steps = self.memory.load_memory(
                constants.INTERMEDIATE_STEPS, []
            )
            self.memory.save_conversation(
                message=user_query, message_type=MessageType.UserMessage
            )

            inputs[constants.CONVERSATION_HISTORY] = deepcopy(
                self.memory.load_conversation()
            )
            inputs[constants.INTERMEDIATE_STEPS] = deepcopy(intermediate_steps)

        return inputs

    def prep_output(
        self,
        inputs: Dict[str, str],
        output: AgentFinish,
        return_only_outputs: bool = False,
    ) -> Dict[str, Any]:
        """Save conversation into memory and prep outputs."""
        output_dict = output.format_output()
        if self.memory is not None:
            self.memory.save_conversation(
                message=output.message, message_type=MessageType.AIMessage
            )
            self.memory.save_memory(
                key=constants.INTERMEDIATE_STEPS, value=output.intermediate_steps
            )

        if return_only_outputs:
            return output_dict
        else:
            return {**inputs, **output_dict}

    def run(
        self,
        user_query: str,
        return_only_outputs: bool = False,
    ) -> Dict[str, Any]:
        """Wrapper for _run function by formatting the input and outputs

        Args:
            user_query: user query
            return_only_outputs: boolean for whether to return only outputs in the
                response. If True, only new keys generated by this chain will be
                returned. If False, both input keys and new keys generated by this
                chain will be returned. Defaults to False.

        """
        inputs = self.prep_inputs(user_query)
        logger.info(f"\n Input to agent: {inputs}")
        try:
            output = self._run(inputs)
        except (KeyboardInterrupt, Exception) as e:
            raise e

        return self.prep_output(inputs, output, return_only_outputs)

    def _run(
        self,
        inputs: Dict[str, Any],
    ) -> AgentFinish:
        """
        Run inputs including user query and past conversation with agent and get response back
        calls take_next_step function to determine what should be the next step after
        collecting all the inputs and memorized contents
        """
        # Construct a mapping of tool name to tool for easy lookup
        name_to_tool_map = {tool.name: tool for tool in self.agent.tools}

        intermediate_steps: List[AgentAction] = inputs[constants.INTERMEDIATE_STEPS]

        # Let's start tracking the number of iterations and time elapsed
        iterations = 0
        time_elapsed = 0.0
        start_time = time.time()
        # We now enter the agent loop (until it returns something).
        while self._should_continue(iterations, time_elapsed):
            logger.info(f"\n Intermediate steps: {intermediate_steps}\n")
            next_step_output = self.should_answer(inputs=inputs)

            # if next_step_output is None which means should ask agent to answer and take next
            # step
            if not next_step_output:
                next_step_output = self.take_next_step(
                    name_to_tool_map,
                    inputs,
                )

            if isinstance(next_step_output, AgentFinish):
                next_step_output.intermediate_steps = intermediate_steps
                return next_step_output

            # stores action output into the conversation as FunctionMessage, which can be used by
            # OpenAIFunctionsAgent
            if isinstance(next_step_output, AgentAction):
                self.memory.save_conversation(
                    message=str(next_step_output.tool_output),
                    name=next_step_output.tool,
                    message_type=MessageType.FunctionMessage,
                )

            intermediate_steps.append(next_step_output)
            # update inputs
            inputs[constants.INTERMEDIATE_STEPS] = intermediate_steps
            inputs[constants.CONVERSATION_HISTORY] = self.memory.load_conversation()

            iterations += 1
            time_elapsed = time.time() - start_time

        # force the termination when shouldn't continue
        output = AgentFinish(
            message="Agent stopped due to iteration limit or time limit.",
            log="",
            intermediate_steps=intermediate_steps,
        )
        return output

    @abstractmethod
    def take_next_step(
        self,
        name_to_tool_map: Dict[str, Tool],
        inputs: Dict[str, str],
    ) -> (AgentFinish, AgentAction):
        """How agent determines the next step after observing the inputs and intermediate
        steps"""

    def _should_continue(self, iterations: int, time_elapsed: float) -> bool:
        if self.max_iterations is not None and iterations >= self.max_iterations:
            return False
        if (
            self.max_execution_time is not None
            and time_elapsed >= self.max_execution_time
        ):
            return False

        return True

    def should_answer(self, inputs) -> Optional[AgentFinish]:
        """
        Let agent determines if it should continue to answer questions
        or that is the end of the conversation
        Args:
            inputs: Dict contains user query and other memorized contents

        Returns:
            None if should answer
            AgentFinish if should NOT answer and respond to user with message
        """
        output = None
        # check if agent should answer this query
        last_query = (
            inputs[constants.CONVERSATION_HISTORY].get_latest_user_message().content
        )
        if self.last_query != last_query:
            output = self.agent.should_answer(**inputs)
            self.last_query = last_query

        return output



from autochain.agent.conversational_agent.conversational_agent import (
    ConversationalAgent,
)
from autochain.chain.chain import Chain
from autochain.memory.buffer_memory import BufferMemory
from autochain.models.chat_openai import ChatOpenAI
from autochain.utils import get_args

# Set logging level
_ = get_args()

prompt = (
    "You are a sales agent who wants to up sale all customer inquire. Your goal is "
    "introducing more expensive options to user"
)

llm = ChatOpenAI(temperature=0)
memory = BufferMemory()
agent = ConversationalAgent.from_llm_and_tools(llm=llm, prompt=prompt)
chain = Chain(agent=agent, memory=memory)

user_query = "How much is this basic rice cooker"
print(f">>> User: {user_query}")
print(
    f""">>> Assistant: 
{chain.run("How much is this basic rice cooker")["message"]}
"""
)



from typing import List, Optional, Dict

from autochain.agent.conversational_agent.conversational_agent import (
    ConversationalAgent,
)

from autochain.agent.structs import AgentOutputParser
from autochain.agent.message import BaseMessage
from autochain.chain.chain import Chain
from autochain.memory.base import BaseMemory
from autochain.memory.buffer_memory import BufferMemory
from autochain.models.base import BaseLanguageModel
from autochain.models.chat_openai import ChatOpenAI
from autochain.tools.base import Tool


def create_chain_from_test(
    tools: List[Tool],
    memory: Optional[BaseMemory] = None,
    llm: Optional[BaseLanguageModel] = None,
    agent_cls=ConversationalAgent,
    **kwargs
):
    """
    Create Chain for running tests
    Args:
        tools: list of autochain tools
        memory: memory store for chain
        llm: model for agent
        agent_cls: metadata class for instantiating agent
    Returns:
        Chain
    """
    llm = llm or ChatOpenAI(temperature=0)
    memory = memory or BufferMemory()
    agent = agent_cls.from_llm_and_tools(llm=llm, tools=tools, **kwargs)
    return Chain(agent=agent, memory=memory)


def parse_evaluation_response(message: BaseMessage) -> Dict[str, str]:
    """
    Parse the reason and rating from the call to determine if the conversation reaches the
    expected outcome
    """
    response = AgentOutputParser.load_json_output(message)
    return {
        "rating": response.get("rating"),
        "reason": response.get("reason"),
    }



from autochain.agent.conversational_agent.conversational_agent import (
    ConversationalAgent,
)
from autochain.chain.chain import Chain
from autochain.memory.buffer_memory import BufferMemory
from autochain.models.chat_openai import ChatOpenAI
from autochain.utils import get_args

# Set logging level
_ = get_args()

llm = ChatOpenAI(temperature=0)
memory = BufferMemory()
agent = ConversationalAgent.from_llm_and_tools(llm=llm)
chain = Chain(agent=agent, memory=memory)

user_query = "Write me a poem about AI"
print(f">> User: {user_query}")
print(
    f""">>> Assistant: 
{chain.run(user_query)["message"]}
"""
)



from __future__ import annotations

import logging
from abc import abstractmethod
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from pydantic import Extra, Field, BaseModel
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from autochain.agent.message import BaseMessage
from autochain.tools.base import Tool

logger = logging.getLogger(__name__)


class Generation(BaseModel):
    """Output of a single generation."""

    message: BaseMessage
    """Generated text output."""

    generation_info: Optional[Dict[str, Any]] = None
    """Raw generation info response from the provider"""
    """May include things like reason for finishing (e.g. in OpenAI)"""
    # TODO: add log probs


class LLMResult(BaseModel):
    """Class that contains all relevant information for an LLM Result."""

    generations: List[Generation]
    """List of the things generated. This is List[List[]] because
    each input could have multiple generations."""
    llm_output: Optional[dict] = None
    """For arbitrary LLM provider specific output."""


class EmbeddingResult(BaseModel):
    texts: List[str]
    embeddings: List[List[float]]


class BaseLanguageModel(BaseModel):
    """Wrapper around OpenAI Chat large language models.

    To use, you should have the ``openai`` python package installed, and the
    environment variable ``OPENAI_API_KEY`` set with your API key.

    Any parameters that are valid to be passed to the openai.create call can be passed
    in, even if not explicitly saved on this class.

    Example:
        .. code-block:: python

            from autochain.models import ChatOpenAI
            openai = ChatOpenAI(model_name="gpt-3.5-turbo")
    """

    client: Any  #: :meta private:
    model_name: str = "gpt-3.5-turbo"
    """Model name to use."""
    temperature: float = 0.7
    """What sampling temperature to use."""
    model_kwargs: Dict[str, Any] = Field(default_factory=dict)
    """Holds any model parameters valid for `create` call not explicitly specified."""
    openai_api_key: Optional[str] = None
    openai_organization: Optional[str] = None
    request_timeout: Optional[Union[float, Tuple[float, float]]] = None
    """Timeout for requests to OpenAI completion API. Default is 600 seconds."""
    max_retries: int = 6
    """Maximum number of retries to make when generating."""
    n: int = 1
    """Number of chat completions to generate for each prompt."""
    max_tokens: Optional[int] = None
    """Maximum number of tokens to generate."""

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.ignore

    @property
    def _default_params(self) -> Dict[str, Any]:
        """Get the default parameters for calling OpenAI API."""
        return {
            "model": self.model_name,
            "request_timeout": self.request_timeout,
            "max_tokens": self.max_tokens,
            "n": self.n,
            "temperature": self.temperature,
            **self.model_kwargs,
        }

    def _create_retry_decorator(self) -> Callable[[Any], Any]:
        import openai

        min_seconds = 1
        max_seconds = 60
        # Wait 2^x * 1 second between each retry starting with
        # 4 seconds, then up to 10 seconds, then 10 seconds afterwards
        return retry(
            reraise=True,
            stop=stop_after_attempt(self.max_retries),
            wait=wait_exponential(multiplier=1, min=min_seconds, max=max_seconds),
            retry=(
                retry_if_exception_type(openai.error.Timeout)
                | retry_if_exception_type(openai.error.APIError)
                | retry_if_exception_type(openai.error.APIConnectionError)
                | retry_if_exception_type(openai.error.RateLimitError)
                | retry_if_exception_type(openai.error.ServiceUnavailableError)
            ),
            before_sleep=before_sleep_log(logger, logging.WARNING),
        )

    def generate_with_retry(self, **kwargs: Any) -> Any:
        """Use tenacity to retry the completion call."""
        retry_decorator = self._create_retry_decorator()

        @retry_decorator
        def _generate_with_retry(**kwargs: Any) -> Any:
            return self.client.create(**kwargs)

        return _generate_with_retry(**kwargs)

    @abstractmethod
    def generate(
        self,
        messages: List[BaseMessage],
        functions: Optional[List[Tool]] = None,
        stop: Optional[List[str]] = None,
    ) -> LLMResult:
        pass

    def encode(self, texts: List[str]) -> EmbeddingResult:
        pass



from autochain.tools.base import Tool
from autochain.tools.google_search.util import GoogleSearchAPIWrapper


class GoogleSearchTool(Tool):
    """Tool that has capability to query the Google Search API and get back json."""

    name = "Google Search Results JSON"
    description = (
        "A wrapper around Google Search. "
        "Useful for when you need to answer questions about current events. "
        "Input should be a search query. Output is a JSON array of the query results"
    )
    num_results: int = 4
    api_wrapper: GoogleSearchAPIWrapper

    def _run(
        self,
        query: str,
    ) -> str:
        """Use the tool."""
        return str(self.api_wrapper.results(query, self.num_results))



from autochain.chain.chain import Chain
from autochain.memory.buffer_memory import BufferMemory
from autochain.models.chat_openai import ChatOpenAI
from autochain.tools.base import Tool
from autochain.agent.conversational_agent.conversational_agent import (
    ConversationalAgent,
)
from autochain.utils import get_args

# Set logging level
_ = get_args()

llm = ChatOpenAI(temperature=0)
tools = [
    Tool(
        name="Get weather",
        func=lambda *args, **kwargs: "Today is a sunny day",
        description="""This function returns the weather information""",
    )
]

memory = BufferMemory()
agent = ConversationalAgent.from_llm_and_tools(llm=llm, tools=tools)
chain = Chain(agent=agent, memory=memory)

user_query = "what is the weather today"
print(f">> User: {user_query}")
print(f">> Assistant: {chain.run(user_query)['message']}")
next_user_query = "Boston"
print(f">> User: {next_user_query}")
print(f">> Assistant: {chain.run(next_user_query)['message']}")



from __future__ import annotations

PLANNING_PROMPT_TEMPLATE = """You are an assistant who tries to have helpful conversation 
with user based on previous conversation and previous tools outputs from tools. 
${prompt}
Use tool when provided. If there is no tool available, respond with have a helpful and polite 
conversation.

Assistant has access to the following tools:
${tools}

Previous conversation so far:
${history}

Previous tools outputs:
${agent_scratchpad}

Please respond user question in JSON format as described below
RESPONSE FORMAT:
{
  "thoughts": {
    "plan": "Given workflow policy and previous tools outputs, what is the next step after the previous conversation",
    "need_use_tool": "answer with 'Yes' if needs to use another tool not previously used else 'No'"
  },
  "tool": {
    "name": "tool name, should be one of [${tool_names}] or empty if tool is not needed",
    "args": {
      "arg_name": "arg value from conversation history or tools outputs to run tool"
    }
  },
  "response": "clarifying required args for that tool or response to user.",
}

Ensure the response can be parsed by Python json.loads
"""

SHOULD_ANSWER_PROMPT_TEMPLATE = """You are a support agent. 
Given the following conversation so far, has user acknowledged question is resolved, 
such as thank you or that's all. 
Answer with yes or no.

Conversation:
${history}
"""

FIX_TOOL_INPUT_PROMPT_TEMPLATE = """Tool have the following spec and input provided
Spec: "{tool_description}"
Inputs: "{inputs}"
Running this tool failed with the following error: "{error}"
What is the correct input in JSON format for this tool?
"""


CLARIFYING_QUESTION_PROMPT_TEMPLATE = """You are a support agent who is going to use '${tool_name}' tool.
Check if you have enough information from the previous conversation and tools outputs to use tool based on the spec below.
"${tool_desp}"

Previous conversation so far:
${history}

Previous tools outputs:
${agent_scratchpad}

Please respond user question in JSON format as described below
RESPONSE FORMAT:
{
    "has_arg_value": "Do values for all input args for '${tool_name}' tool exist? answer with Yes or No",
    "clarifying_question": "clarifying question to user to ask for missing information"
}
Ensure the response can be parsed by Python json.loads"""



from __future__ import annotations

from abc import ABC, abstractmethod
from string import Template
from typing import Any, List, Optional, Sequence, Union

from pydantic import BaseModel, Extra

from autochain.agent.message import ChatMessageHistory
from autochain.agent.prompt_formatter import JSONPromptTemplate
from autochain.agent.structs import AgentAction, AgentFinish, AgentOutputParser
from autochain.models.base import BaseLanguageModel
from autochain.tools.base import Tool


class BaseAgent(BaseModel, ABC):
    output_parser: AgentOutputParser = None
    llm: BaseLanguageModel = None
    tools: Sequence[Tool] = []

    @classmethod
    def from_llm_and_tools(
        cls,
        llm: BaseLanguageModel,
        tools: Sequence[Tool],
        prompt: str,
        output_parser: Optional[AgentOutputParser] = None,
        input_variables: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> BaseAgent:
        """Construct an agent from an LLM and tools."""

    def should_answer(
        self, should_answer_prompt_template: str = "", **kwargs
    ) -> Optional[AgentFinish]:
        """Determine if agent should continue to answer user questions based on the latest user
        query"""
        return None

    @abstractmethod
    def plan(
        self,
        history: ChatMessageHistory,
        intermediate_steps: List[AgentAction],
        **kwargs: Any,
    ) -> Union[AgentAction, AgentFinish]:
        """
        Plan the next step. either taking an action with AgentAction or respond to user with AgentFinish
        Args:
            history: entire conversation history between user and agent including the latest query
            intermediate_steps: List of AgentAction that has been performed with outputs
            **kwargs: key value pairs from chain, which contains query and other stored memories

        Returns:
            AgentAction or AgentFinish
        """

    def clarify_args_for_agent_action(
        self,
        agent_action: AgentAction,
        history: ChatMessageHistory,
        intermediate_steps: List[AgentAction],
        **kwargs: Any,
    ) -> Union[AgentAction, AgentFinish]:
        """
        Ask clarifying question if needed. When agent is about to perform an action, we could
        use this function with different prompt to ask clarifying question for input if needed.
        Sometimes the planning response would already have the clarifying question, but we found
        it is more precise if there is a different prompt just for clarifying args

        Args:
            agent_action: agent action about to take
            history: conversation history including the latest query
            intermediate_steps: list of agent action taken so far
            **kwargs:

        Returns:
            Either a clarifying question (AgentFinish) or take the planned action (AgentAction)
        """
        return agent_action

    def fix_action_input(
        self, tool: Tool, action: AgentAction, error: str
    ) -> Optional[AgentAction]:
        """If the tool failed due to error, what should be the fix for inputs"""
        pass

    @staticmethod
    def get_prompt_template(
        prompt: str = "",
        input_variables: Optional[List[str]] = None,
    ) -> JSONPromptTemplate:
        """Create prompt in the style of the zero shot agent.

        Args:
            prompt: message to be injected between prefix and suffix.
            input_variables: List of input variables the final prompt will expect.

        Returns:
            A PromptTemplate with the template assembled from the pieces here.
        """
        template = Template(prompt)

        if input_variables is None:
            input_variables = ["input", "agent_scratchpad"]
        return JSONPromptTemplate(template=template, input_variables=input_variables)



"""Default implementation of Chain"""
import logging
from typing import Dict

from autochain.agent.message import MessageType
from autochain.agent.structs import AgentAction, AgentFinish
from autochain.chain.base_chain import BaseChain
from autochain.errors import ToolRunningError
from autochain.tools.base import Tool
from autochain.tools.simple_handoff.tool import HandOffToAgent

logger = logging.getLogger(__name__)


class Chain(BaseChain):
    """
    Default chain with take_next_step implemented
    It handles a few common error cases with agent, such as taking repeated action with same
    inputs and whether agent should continue the conversation
    """

    return_intermediate_steps: bool = False
    handle_parsing_errors = True
    graceful_exit_tool: Tool = HandOffToAgent()

    def handle_repeated_action(self, agent_action: AgentAction) -> AgentFinish:
        if agent_action.model_response:
            print(
                f"Action taken before: {agent_action.tool}, "
                f"input: {agent_action.tool_input}"
            )
            return AgentFinish(
                message=agent_action.response,
                log=f"Action taken before: {agent_action.tool}, "
                f"input: {agent_action.tool_input}",
            )
        else:
            return AgentFinish(
                message=self.graceful_exit_tool.run(),
                log=f"Gracefully exit due to repeated action",
            )

    def take_next_step(
        self,
        name_to_tool_map: Dict[str, Tool],
        inputs: Dict[str, str],
    ) -> (AgentFinish, AgentAction):
        """
        How agent determines the next step after observing the inputs and intermediate steps
        Args:
            name_to_tool_map: map of tool name to the actual tool object
            inputs: a dictionary of all inputs, such as user query, past conversation and
                tools outputs

        Returns:
            Either AgentFinish to respond to user or AgentAction to take the next action
        """

        try:
            # Call the LLM to see what to do.
            output = self.agent.plan(
                **inputs,
            )
        except Exception as e:
            if not self.handle_parsing_errors:
                raise e
            tool_output = f"Invalid or incomplete response due to {e}"
            print(tool_output)
            output = AgentFinish(message=self.graceful_exit_tool.run(), log=tool_output)
            return output

        if isinstance(output, AgentAction):
            output = self.agent.clarify_args_for_agent_action(output, **inputs)

        # If agent plans to respond to AgentFinish or there is a clarifying question, respond to
        # user by returning AgentFinish
        if isinstance(output, AgentFinish):
            return output

        if isinstance(output, AgentAction):
            tool_output = ""
            # Check if tool is supported
            if output.tool in name_to_tool_map:
                tool = name_to_tool_map[output.tool]

                # how to handle the case where same action with same input is taken before
                if output.tool_input == self.memory.load_memory(tool.name):
                    return self.handle_repeated_action(output)

                self.memory.save_memory(tool.name, output.tool_input)
                # We then call the tool on the tool input to get an tool_output
                try:
                    tool_output = tool.run(output.tool_input)
                except ToolRunningError as e:
                    new_agent_action = self.agent.fix_action_input(
                        tool, output, error=str(e)
                    )
                    if (
                        new_agent_action
                        and new_agent_action.tool_input != output.tool_input
                    ):
                        tool_output = tool.run(output.tool_input)

                print(
                    f"Took action '{tool.name}' with inputs '{output.tool_input}', "
                    f"and the tool_output is {tool_output}"
                )
            else:
                tool_output = f"Tool {output.tool} if not supported"

            output.tool_output = tool_output
            return output
        else:
            raise ValueError(f"Unsupported action: {type(output)}")



import json

from autochain.agent.openai_funtions_agent.openai_functions_agent import (
    OpenAIFunctionsAgent,
)
from autochain.models.chat_openai import ChatOpenAI
from autochain.tools.base import Tool
from autochain.workflows_evaluation.base_test import BaseTest, TestCase, WorkflowTester
from autochain.workflows_evaluation.test_utils import (
    create_chain_from_test,
)
from autochain.utils import get_args


def get_current_weather(location: str, unit: str = "fahrenheit"):
    """Get the current weather in a given location"""
    weather_info = {
        "location": location,
        "temperature": "72",
        "unit": unit,
        "forecast": ["sunny", "windy"],
    }
    return json.dumps(weather_info)


class TestGetWeatherWithFunctionCalling(BaseTest):
    prompt = """You are a weather support agent tries to get weather information for requested 
user location"""

    tools = [
        Tool(
            name="get_current_weather",
            func=get_current_weather,
            description="""Get the current weather in a given location""",
        )
    ]

    test_cases = [
        TestCase(
            test_name="get weather for boston",
            user_context="want to get current weather information; location in Boston",
            expected_outcome="found weather information in Boston",
        ),
    ]

    llm = ChatOpenAI(temperature=0)
    chain = create_chain_from_test(
        tools=tools, agent_cls=OpenAIFunctionsAgent, llm=llm, prompt=prompt
    )


if __name__ == "__main__":
    tester = WorkflowTester(
        tests=[TestGetWeatherWithFunctionCalling()],
        output_dir="./test_results",
    )

    args = get_args()
    if args.interact:
        tester.run_interactive()
    else:
        tester.run_all_tests()
